/* Transparent socks proxy.
    Copyright (C) 1999  V. Ganesh <ganesh@vxindia.veritas.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <pwd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>

#include <socks.h>

/* paranoia, I know. the socks libraries should take care of 
   non-socks sockets, but ...
*/
#define real_listen		listen
#define real_bind		bind
#define real_accept		accept
#define real_connect		connect
#define real_getsockname	getsockname
#define real_read		read
#define real_write		write
#define real_close		close
#define real_shutdown		shutdown

#define MAXTRIES		3
#define SLEEP_BETWEEN_TRIES	2
#define BUFSIZE			16384
#define CONNECT_TIMEOUT		60
#define DATA_TIMEOUT		300

void socks_serve(int);
void pass_data(int, int);
void alarm_handler(int);

static void logmsg(char *);
static void logerr(char *);

int isdaemon = 1;
short bindport = 1211;

main(argc, argv)
	int	argc;
	char	**argv;
{
	struct sockaddr_in	addr;
	int			addrlen;
	int			listenfd, newfd;
	int			newpid;
	char			c;

	while ((c = getopt(argc, argv, "fp:")) != EOF) {
		switch(c) {
		case 'f':
			isdaemon = 0;
			break;
		case 'p':
			if (!atoi(optarg)) {
				fprintf(stderr, "Invalid port %s\n", optarg);
				exit(-1);
			}
			bindport = atoi(optarg);
			break;
		case '?':
			fprintf(stderr, "Usage: %s [-f] [-p port]\n", argv[0]);
			exit(1);
			break;
		}

	}

	if (isdaemon) {
		if ((newpid = fork()) < 0) {
			perror("fork");
			exit(-1);
		}
		if (newpid) { /* parent */
			exit(0);
		}
		setsid();
		setpgrp();
		openlog("transocks", LOG_PID, LOG_DAEMON);
	}

	listenfd = socket(AF_INET, SOCK_STREAM, 0);
	if (listenfd < 0) {
		logerr("socket");
		exit(-1);
	}

	bzero(&addr, sizeof (addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(bindport);

	if (real_bind(listenfd, (struct sockaddr *)&addr, sizeof (addr)) < 0) {
		logerr("bind");
		exit(-1);
	}

	if (real_listen(listenfd, SOMAXCONN) < 0) {
		logerr("listen");
		exit(-1);
	}

	signal(SIGCHLD, SIG_IGN);

	for (;;) {
		if ((newfd = real_accept(listenfd, (struct sockaddr *)&addr, &addrlen)) < 0) {
			logerr("accept");
			exit(-1);
		}

		socks_serve(newfd);
		real_close(newfd);
		continue;
		newpid = fork();
		if (newpid == 0) {
			socks_serve(newfd);
			exit(0);
		} else if (newpid < 0) {
			logerr("fork");
			exit(-1);
		}

		real_close(newfd);
	}
}

void
socks_serve(newfd)
	int	newfd;
{
	struct sockaddr_in	dstaddr;
	int			dstlen;
	int			tries = MAXTRIES;
	int			clientfd, serverfd;
	char			msg[1024];

	if (isdaemon) {
		closelog();
		openlog("transocks", LOG_PID, LOG_DAEMON);
	}

	clientfd = newfd;
	serverfd = socket(AF_INET, SOCK_STREAM, 0);
	if (serverfd < 0) {
		logerr("socket");
		exit(-1);
	}
	
	signal(SIGHUP, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, alarm_handler);
	alarm(CONNECT_TIMEOUT);

	dstlen = sizeof(dstaddr);
	real_getsockname(clientfd, (struct sockaddr *)&dstaddr, &dstlen);
	sprintf(msg, "request to: %s port %d", inet_ntoa(dstaddr.sin_addr), ntohs(dstaddr.sin_port));
	logmsg(msg);

	/* my socks server has a tendency to refuse connections, maybe because of load.
	   irritating, so we retry after a few seconds
	*/
	while (tries--) {
		if (connect(serverfd, (struct sockaddr *)&dstaddr, sizeof (dstaddr)) == 0) {
			alarm(0);
			pass_data(clientfd, serverfd);
			return;
		}
		sleep(SLEEP_BETWEEN_TRIES);
	}
	logerr("connect");
	return;
	exit(-1);
}

void
pass_data(clientfd, serverfd)
	int	clientfd;
	int	serverfd;
{
	char	buf[BUFSIZE];
	int	buflen;
	fd_set	readfds;
	int	fdmax;
	int	close_state = 0;
	int	nwritten;
	char	*bp;

#define C_TO_S 1
#define S_TO_C 2

	fdmax = (clientfd > serverfd) ? clientfd : serverfd;
	fdmax++;

	/* a possibly misguided attempt is made to implement
	   shutdown semantics. seems to be harmless though
	*/
	while (1) {
		FD_ZERO(&readfds);
		FD_SET(clientfd, &readfds);
		FD_SET(serverfd, &readfds);

		alarm(DATA_TIMEOUT);
		if (select(fdmax, &readfds, NULL, NULL, NULL) < 0) {
			logerr("select");
			break;
		}

		alarm(0);
		if (FD_ISSET(clientfd, &readfds)) {
			buflen = real_read(clientfd, buf, BUFSIZE);
			if (buflen < 0 && errno != EPIPE) {
				logerr("read from client");
				break;
			}
			if (buflen <= 0) {
				shutdown(serverfd, 1);
				close_state |= C_TO_S;
				if (close_state == (C_TO_S | S_TO_C)) {
					break;
				}
			}
			bp = buf;
			while (buflen) {
				nwritten = write(serverfd, bp, buflen);
				if (nwritten < 0 && errno != EPIPE) {
					logerr("write to server");
					break;
				}
				if (nwritten <= 0) {
					real_shutdown(clientfd, 0);
					close_state |= C_TO_S;
					if (close_state == (C_TO_S | S_TO_C)) {
						break;
					}
				}
				buflen -= nwritten;
				bp += nwritten;
			}
		}
		if (FD_ISSET(serverfd, &readfds)) {
			buflen = read(serverfd, buf, BUFSIZE);
			if (buflen < 0 && errno != EPIPE) {
				logerr("read from server");
				break;
			}
			if (buflen <= 0) {
				real_shutdown(clientfd, 1);
				close_state |= S_TO_C;
				if (close_state == (C_TO_S | S_TO_C)) {
					break;
				}
			}
			bp = buf;
			while (buflen) {
				nwritten = real_write(clientfd, bp, buflen);
				if (nwritten < 0 && errno != EPIPE) {
					logerr("write to client");
					break;
				}
				if (nwritten <= 0) {
					shutdown(serverfd, 0);
					close_state |= S_TO_C;
					if (close_state == (C_TO_S | S_TO_C)) {
						break;
					}
				}
				buflen -= nwritten;
				bp += nwritten;
			}
		}
	}
}

void
alarm_handler(sig)
	int	sig;
{
	exit(0);
}

static void
logmsg(char *msg)
{
	if (isdaemon) {
		syslog(LOG_ERR, msg);
		return;
	}
	fprintf(stderr, "transocks[%d]: %s\n", getpid(), msg);
}

static void
logerr(char *msg)
{
	char	tmp[2048];

	sprintf(tmp, "%s: %s", msg, strerror(errno));
	logmsg(tmp);
}
